@@ -1,17 +1,19 @@
-from .classifier import (MLP, SVM, AdaBoost, Keras, RandomForest, Transformer,
-                         XGBoost)
+from .classifier import (MLP, RNN, SVM, AdaBoost, Keras, RandomForest,
+                         Transformer, XGBoost)
 from .embedding import CountVectorizer, Doc2Vec, OneHot, Sequence, TfIdf
 from .linguistics import (Contract, FindAll, Lemmatizer, Phrase, Snowball, Sub,
                           Tokenize)
 from .multi import Merge
 from .preprocessing import Normalize
 from .reduction import NMF, PCA, SVD, UMAP
-from .special import SplitHashtag
+from .regressor import KerasRegressor, MLPRegressor
+from .special import DomainName, Lambda, SplitHashtag
 
 __all__ = [
     "Tokenize", "Snowball", "Phrase", "Sub", "Contract", "FindAll",
     "TfIdf", "CountVectorizer", "Sequence", "OneHot", "Doc2Vec",
     "SVD", "PCA", "NMF", "UMAP", "Lemmatizer",
     "RandomForest", "AdaBoost", "SVM", "MLP", "Keras", "Transformer",
-    "XGBoost", "Merge", "Normalize", "SplitHashtag"
+    "XGBoost", "Merge", "Normalize", "SplitHashtag",
+    "MLPRegressor", "KerasRegressor", "Lambda", "RNN", "DomainName"
 ]
